# AUTHOR: Louis Tsiattalou
# DESCRIPTION: Test matcher module. Execute with `python -m unittest discover -s test/`

import re
import unittest

import pandas as pd

from tfidf_matcher.matcher import matcher
from tfidf_matcher.ngrams import ngrams


class TestMatcher(unittest.TestCase):

    test_list = ["hello", "theseare", "definitely", "companynames"]
    test_lookup = ["hullo", "company", "finite", "cesare"]

    max_k_matches = len(test_lookup) - 1

    def test_matcher_output_shape(self):
        """Test that the output of matcher is as expected."""
        for k in range(1, self.max_k_matches + 1):
            self.assertEqual(
                matcher(self.test_list, self.test_lookup, k_matches=k).shape,
                (len(self.test_list), (k * 3) + 1),
            )

    def test_matcher_outputs_found_in_original(self):
        """Test that all of the matched strings are found in the original."""
        res = matcher(self.test_list, self.test_lookup, k_matches=self.max_k_matches)[
            "Original Name"
        ]
        self.assertTrue(all([x in self.test_list for x in res]))

    def test_matcher_outputs_found_in_lookup(self):
        """Test that all of the matcher lookup outputs are found in the lookups list."""
        res = matcher(self.test_list, self.test_lookup, k_matches=self.max_k_matches)
        lookup_cols = [c for c in res.columns if re.match(r"^Lookup \d+$", c)]
        lookup_unique_values = pd.unique(res[lookup_cols].values.ravel("K"))
        self.assertTrue(all([x in self.test_lookup for x in lookup_unique_values]))

    def test_matcher_scores_normalized(self):
        """Test that the matcher matchscores are between 0 and 1."""
        res = matcher(self.test_list, self.test_lookup, k_matches=self.max_k_matches)
        confidence_cols = [
            c for c in res.columns if re.match(r"^Lookup \d+ Confidence$", c)
        ]
        for confidence_col in confidence_cols:
            self.assertTrue(all([0 <= x <= 1 for x in res[confidence_col]]))


if __name__ == "__main__":
    unittest.main()



# AUTHOR: Louis Tsiattalou
# DESCRIPTION: Extract ngrams from a string.

import re


def ngrams(string, n=3):
    """Generate a full list of ngrams from a list of strings

    :param string: List of strings to generate ngrams from.
    :type string: list (of strings)
    :param n: Maximum length of the n-gram. Defaults to 3.
    :type n: int
    :raises AssertionError: If you pass in a list that contains datatypes other than `string`, you're in trouble!
    :return: Returns list of ngrams generated from the input string.
    :rtype: list
    """

    # Assert string type
    assert type(string) == type("string"), "String not passed in!"

    # Remove Punctuation from the string
    string = re.sub(r"[,-./]|\sBD", r"", string)

    # Generate zip of ngrams (n defined in function argument)
    ngrams = zip(*[string[i:] for i in range(n)])

    # Return ngram list
    return ["".join(ngram) for ngram in ngrams]



# AUTHOR: Louis Tsiattalou
# DESCRIPTION: Match list items to closest tf-idf match in second list.

import pandas as pd
from tfidf_matcher.ngrams import ngrams
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.neighbors import NearestNeighbors


def matcher(original=[], lookup=[], k_matches=5, ngram_length=3):
    """Takes two lists, returns top `k` matches from `lookup` dataset.

    This function does this by:
    - Splitting the `lookup` list into ngrams.
    - Transforming the resulting ngram list into a TF-IDF Sparse Matrix.
    - Fit a NearestNeighbours Model to the matrix using the lookup data.
    - Transform the `original` list into a TF-IDF Sparse Matrix.
    - Calculates distances to all the `n-matches` nearest neighbours
    - Then extract the `original`, `n-matches` closest lookups, and calculate
    a match score (abs(1 - Distance to Nearest Neighbour))

    :param original: List of strings to generate ngrams from.
    :type original: list (of strings), or Pandas Series.
    :param lookup: List of strings to match against.
    :type lookup: list (of strings), or Pandas Series.
    :param k_matches: Number of matches to return.
    :type k_matches: int
    :param ngram_length: Length of Ngrams returned by `tfidf_matcher.ngrams` callable
    :type ngram_length: int
    :raises AssertionError: Throws an error if the datatypes in `original` aren't strings.
    :raises AssertionError: Throws an error if the datatypes in `lookup` aren't strings.
    :raises AssertionError: Throws an error if `k_matches` isn't an integer.
    :raises AssertionError: Throws an error if k_matches > len(lookup)
    :raises AssertionError: Throws an error if ngram_length isn't an integer
    :return: Returns a Pandas dataframe with the `original` list,
        `k_matches` columns containing the closest matches from `lookup`,
        as well as a Match Score for the closest of these matches.
    :rtype: Pandas dataframe
    """

    # Assertions
    assert all(
        [type(x) == type("string") for x in original]
    ), "Original contains non-str elements!"
    assert all(
        [type(x) == type("string") for x in lookup]
    ), "Lookup contains non-str elements!"
    assert type(k_matches) == type(0), "k_matches must be an integer"
    assert k_matches < len(
        lookup
    ), "k_matches must be shorter than the total length of the lookup list"
    assert type(ngram_length) == type(0), "ngram_length must be an integer"

    # Enforce listtype, set to lower
    original = list(original)
    lookup = list(lookup)
    original_lower = [x.lower() for x in original]
    lookup_lower = [x.lower() for x in lookup]

    # Set ngram length for TfidfVectorizer callable
    def ngrams_user(string, n=ngram_length):
        return ngrams(string, n)

    # Generate Sparse TFIDF matrix from Lookup corpus
    vectorizer = TfidfVectorizer(min_df=1, analyzer=ngrams_user)
    tf_idf_lookup = vectorizer.fit_transform(lookup_lower)

    # Fit KNN model to sparse TFIDF matrix generated from Lookup
    nbrs = NearestNeighbors(n_neighbors=k_matches, n_jobs=-1, metric="cosine").fit(
        tf_idf_lookup
    )

    # Use nbrs model to obtain nearest matches in lookup dataset. Vectorize first.
    tf_idf_original = vectorizer.transform(original_lower)
    distances, lookup_indices = nbrs.kneighbors(tf_idf_original)

    # Extract top Match Score (which is just the distance to the nearest neighbour),
    # Original match item, and Lookup matches.
    original_name_list = []
    confidence_list = []
    index_list = []
    lookup_list = []
    # i is 0:len(original), j is list of lists of matches
    for i, lookup_index in enumerate(lookup_indices):
        original_name = original[i]
        # lookup names in lookup list
        lookups = [lookup[index] for index in lookup_index]
        # transform distances to confidences and store
        confidence = [1 - round(dist, 2) for dist in distances[i]]
        original_name_list.append(original_name)
        # store index
        index_list.append(lookup_index)
        confidence_list.append(confidence)
        lookup_list.append(lookups)

    # Convert to df
    df_orig_name = pd.DataFrame(original_name_list, columns=["Original Name"])

    df_lookups = pd.DataFrame(
        lookup_list, columns=["Lookup " + str(x + 1) for x in range(0, k_matches)]
    )
    df_confidence = pd.DataFrame(
        confidence_list,
        columns=["Lookup " + str(x + 1) + " Confidence" for x in range(0, k_matches)],
    )
    df_index = pd.DataFrame(
        index_list,
        columns=["Lookup " + str(x + 1) + " Index" for x in range(0, k_matches)],
    )

    # bind columns
    matches = pd.concat([df_orig_name, df_lookups, df_confidence, df_index], axis=1)

    # reorder columns | can be skipped
    lookup_cols = list(matches.columns.values)
    lookup_cols_reordered = [lookup_cols[0]]
    for i in range(1, k_matches + 1):
        lookup_cols_reordered.append(lookup_cols[i])
        lookup_cols_reordered.append(lookup_cols[i + k_matches])
        lookup_cols_reordered.append(lookup_cols[i + 2 * k_matches])
    matches = matches[lookup_cols_reordered]
    return matches



#!/usr/bin/env python

from setuptools import setup, find_packages
long_description_md = open("README.md", encoding = "utf-8").read()

setup(
    name="tfidf_matcher",
    version="0.3.0",
    author="Louis Tsiattalou",
    author_email="louis.tsi@gmail.com",
    description="A small package that enables super-fast TF-IDF based string matching.",
    long_description=long_description_md,
    long_description_content_type="text/markdown",
    url="https://github.com/louistsiattalou/tfidf_matcher",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires='>=3.5',
    install_requires=["scikit-learn", "pandas"],
)



# AUTHOR: Louis Tsiattalou
# DESCRIPTION: Test ngrams module. Execute with `python -m unittest discover -s test/`

import unittest

from tfidf_matcher.ngrams import ngrams


class TestNgrams(unittest.TestCase):
    def test_ngram_count(self):
        """Test that the number of ngrams returned by the function is correct."""
        data = "This is a sentence"
        self.assertEqual(len(ngrams(data, n=1)), 18)
        self.assertEqual(len(ngrams(data, n=2)), 17)
        self.assertEqual(len(ngrams(data, n=3)), 16)
        self.assertEqual(len(ngrams(data, n=18)), 1)

    def test_ngram_length(self):
        """Test that the ngram outputs from the function always contain `n` length ngrams."""
        data = "This is a sentence"
        self.assertTrue(all([len(x) == 1 for x in ngrams(data, n=1)]))
        self.assertTrue(all([len(x) == 2 for x in ngrams(data, n=2)]))
        self.assertTrue(all([len(x) == 3 for x in ngrams(data, n=3)]))
        self.assertTrue(all([len(x) == 18 for x in ngrams(data, n=18)]))


if __name__ == "__main__":
    unittest.main()
